# OBJ - Framework to program bots
#
# obj/base.py
#
# Copyright 2019 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 01-01-2019 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

""" base classes. """

__txt__ = "Framework to progam bots"
__version__ = 98

## imports

import _thread
import collections
import datetime
import logging
import os
import importlib
import json
import sys
import time
import types
import threading

## defines

homedir = os.path.abspath(os.path.expanduser("~"))
logdir = os.path.abspath(os.path.join(homedir, ".objlogs"))
starttime = time.time()

## exceptions

class ENOTPATH(Exception):
    """ path is missing or not correct. """
    pass

class EOVERLOAD(Exception):
    """ overload attempt detected. """
    pass
    
## classes

class Object(object):

    """ The pure object that does load/save from and to JSON files. """

    def __iter__(self):
        return iter(self.__dict__)

    def __repr__(self) -> str:
        """ module.class name repr. """
        return "%s.%s at %s" % (
            self.__class__.__module__,
            self.__class__.__name__,
            str(hex(id(self)))
        )

    def __str__(self):
        """ do a standard, no newlines, dump of this object. """
        return json.dumps(self, default=smooth, sort_keys=True)

    def pretty(self) -> str:
        """ stringify on a 4 indent. """
        return json.dumps(self, default=smooth, indent=4, sort_keys=True)

    def get_attr(self, k, d=""):
        """ get attribute from object. """
        return getattr(self, k, d)

    def load(self, path: str) -> None:
        """ load an object from given path. """
        assert path
        p = os.path.abspath(os.path.join(cfg.workdir, path)) 
        logging.debug("load %s" % path)
        with open(p) as f:
            data = json.load(f)
            for k, v in data.items():
                self.set_attr(k, v)
        if "_container" not in dir(self):
            self._container = {}
        self._container["path"] = path
        return self

    def search(self, m: {}) -> bool:
        """ return whether this object values matches criteria. """ 
        res = False
        for k, v in m.items():
            if k.startswith("_"):
                continue
            vv = self.get(k, None)
            if vv and (v == ""):
                res = True
                continue
            if vv != None:
                if v in str(vv): 
                    res = True
                else:
                    res = False
                    break
        return res

    def save(self, path="", stime="") -> str:
        """ save this object to a file. """
        if not path and not stime:
            try:
                path = self._container["path"]
            except (TypeError, AttributeError, KeyError):
                pass
        if not path:
            path = os.path.join(get_type(self), stime or rtime())
        if not path:
            raise ENOPATH
        if "_container" not in dir(self):
            self._container = {}
        self._container["path"] = path
        logging.info("save %s" % path)
        p = os.path.abspath(os.path.join(cfg.workdir, path)) 
        d, fn = os.path.split(p)
        if not os.path.isdir(d):
            cdir(d)
        with open(p, "w") as f:
            json.dump(self, f, default=smooth, indent=4, sort_keys=True)
        return path

    def set_attr(self, k , v):
        """ set a attribute. """
        return setattr(self, k, v)

class Dotted(Object, dict):

    """ JSON load/saveable dict with dotted access (like a object). """

    def __getattribute__(self, name):
        """ try the dict, try the base class. """
        try:
            return self[name]
        except KeyError as ex:
            return super().__getattribute__(name)

    def __getattr__(self, name):
        """ try the dict, try the base class else if _default set that. """
        try:
            return self[name]
        except KeyError:
            try:
                return super().__getattribute__(name)
            except AttributeError:
                if "_default" in self:
                    self[name] = self["_default"]
        return self.__getattribute__(name)

    def __setattr__(self, key, value):
        """ if method set in base class else set in dict. """
        func = self.get(key)
        if type(func) == types.MethodType:
            raise EOVERLOAD(key)
        if type(value) == types.MethodType:
            return super().__setattr__(key, value)
        else:
            return self.__setitem__(key, value)

    def __setitem__(self, key, value):
        """ check for method overloading. """
        func = self.get(key)
        if type(func) == types.MethodType:
            raise EOVERLOAD(key)
        return super().__setitem__(key, value)

    def last(self, selector={}):
        """ update this object with the latest version in db store. """
        store = Store()
        s = Selector()
        s.update(selector)
        s._types.append(get_type(self))
        tmp = store.last(s)
        if tmp:
            self.update(tmp)

    def format(self, keys):
        """ return string with entries and optional txt. """
        res = ""
        if type(keys) == str:
            keys = [keys, ]
        for k in set(keys):
            if k.startswith("_"):
                continue
            if k in self:
                res += "%s=%s " % (k, str(self[k]))
        return res.rstrip()

    def upgrade(self, d):
        """ only update if value is set. """
        for k, v in d.items():
            if v:
                self[k] = v
        return self

    def upgrade2(self, d):
        """ only update if value is not set. """
        for k, v in self.items():
            if k in d:
                if not v:
                    self[k] = d[k]

class Cache(Dotted):

    ''' cache with class attribute "data" as cache '''

    data = Dotted()

class Cfg(Dotted):

    """ Base Config. """

    def __iter__(self):
        for k in self.keys():
            if k.startswith("_"):
                continue
            yield k

class Register(Dotted):

    """ dict of list of values. """

    def register(self, k: str, v: object) -> None:
        """ register a key, value pair, constructing a list if key is not in object. """
        if k not in self:
            self[k] = []
        self[k].append(v)
        return self[k]

class Selector(Dotted):

    """ collects key=value pairs from a string. """

    def __init__(self, txt=""):
        super().__init__()
        self._dkeys = []
        self._types = []
        if txt:
            self.parse(txt)

    def parse(self, txt):
        """ parse a string into a dict with key=value pairs. """
        txt = bytes(txt, "utf-8").decode("unicode_escape")
        quoted = False
        clsname = ""
        key = ""
        types = []
        value = ""
        words = []
        key2 = ""
        nr = -1
        for word in txt.split():
            nr += 1
            try:
                key, value = word.split("=", 1)
                if nr == 1:
                    if ":" in key:
                        key = key.split(":")[0]
                    types.append(key)
            except (IndexError, ValueError):
                types.append(word)
                if nr == 2:
                    self[word] = ""
                continue
            if value.startswith('"'):
                value = value[1:]
                key2 = key
                quoted = True
            if quoted and '"' in value:
                value = value.replace('"', '')
                quoted = False
                key = key2
            if quoted:
                words.append(value)
                continue
            else:
                words.append(value)
                value = " ".join(words)
            if key and value:
                if value.endswith("-"):
                    value = value[:-1]
                self[key] = value.strip()
                if key not in self._dkeys:
                    self._dkeys.append(key)
        for type in types:
            for m in match_types(type):
               if m not in self._types:
                   self._types.append(m)
        return self

class Store(Dotted):

    """ manages objects stored on disk. """

    def find(self, selector, nodel=True, all=False):
        """ find matching objects module.class:attribute=value """
        for fn in sorted(filenames(selector._types), key=lambda x: fn_time(x)):
            o = construct(fn)
            if not nodel and "_deleted" in o and o._deleted:
                continue
            if o.search(selector):
                yield o
            elif all:
                yield o

    def last(self, selector, nodel=True, all=False):
        """ show last object. """
        if not selector:
            fn = list(filenames(selector._types))[-1]
            return construct(fn)
        fn = sorted(filenames(selector._types), key=lambda x: fn_time(x))
        if fn:
            return construct(fn[-1])

## instances

cfg = Cfg(_default="")
cfg.workdir = ""

## utility functions

def cdir(path) -> bool:
    """ create directory. """
    res = ""
    for p in path.split(os.sep):
        res += "%s%s" % (p, os.sep)
        p = os.path.abspath(os.path.normpath(res))
        try:
            os.mkdir(p)
        except (IsADirectoryError, NotADirectoryError, FileExistsError):
            pass
    return True

def construct(fn):
    """ construct a type and initialize it with data JSON dict. """
    cls = get_cls(get_fntype(fn))
    return cls().load(fn)
    
def filenames(types=[]):
    """ show file names of objects stored on path (index into the workdir). """
    global cfg
    res = []
    for type in types:
        p = os.path.abspath(os.path.join(cfg.workdir, type or ""))
        for rootdir, dirs, files in os.walk(p, topdown=True):
            if p not in rootdir:
                continue
            for fn in files:
                yield os.path.join(rootdir, fn).split(cfg.workdir)[-1][1:]

def fn_time(daystr: str) -> None:
    """ return the time determined from the filename of an object. """
    daystr = daystr.replace("_", ":")
    datestr = " ".join(daystr.split(os.sep)[-2:])
    datestr = datestr.split(".")[0]
    try:
        t = time.mktime(time.strptime(datestr, "%Y-%m-%d %H:%M:%S"))
    except ValueError:
        t = 0
    return t

def get_cls(name):
    """ return class from sys.modules. """
    name, c = name.rsplit(".", 1)
    mod = importlib.import_module(name)
    return getattr(mod, c)

def get_type(obj):
    """ return this type of object. """
    return obj.__class__.__module__ + "." + obj.__class__.__qualname__

def get_types():
    """ return all types available on disk. """
    return sorted(os.listdir(cfg.workdir))

def get_fntype(fn):
    """ return class from filename. """
    return fn.split(os.sep)[0]

def locked(func, *args, **kwargs):

    """ lock a method or function. """

    lock = _thread.allocate_lock()

    def lockedfunc(*args, **kwargs):
        """ encapsulate the function into a blocking one. """
        lock.acquire()
        res = None
        try:
            res = func(*args, **kwargs)
        finally:
            try:
                lock.release()
            except:
                pass
        return res
    return lockedfunc

def match_types(name):
    """ search for natching types. """
    if name == None:
        return []
    names = []
    for t in get_types():
        if name.lower() in t.lower():
            names.append(t)
    return names

def prefixes():
    """ show list of subdirectories in the workdir. """
    return sorted(set([x.split(".")[-1].lower() for x in os.listdir(cfg.workdir)]))

def rtime():
    """ return filename base of current time. """ 
    return str(datetime.datetime.now()).replace(" ", os.sep)

def slice(o, keys=None):
    """ return subset of an object. """
    cls = get_cls(get_type(o))
    oo = cls()
    if not keys:
        keys = o.keys()
    for k in keys:
        if k.startswith("_"):
            continue
        try:
            oo[k] = o[k]
        except KeyError:
            pass
    return oo 

def smooth(o):
    """ deal with objects that are not convertable to basic JSON types. """
    if isinstance(o, Object):
        return vars(o)
    if type(o) not in [dict, list, str, int, float, bool, None]:
        return repr(o)
    return str(o)
