# OBJ - Framework to program bots
#
# obj/bot.py
#
# Copyright 2019 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 01-01-2019 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

"""
    bot base class.
    
"""

__version__ = 1

## imports

from obj.base import Dotted, cfg, homedir
from obj.event import Event
from obj.fleet import Fleet
from obj.handler import Worker

import os
import sys
import time

## defines

class Cfg(Dotted):

    """ Config class. """

    pass

## classes

class Event(Event):

    """ Shell event. """

    pass

## classes

class OutputCache(Dotted):

    """ cache used to cache output to bots and other output objects. """

    def add(self, dest, txt):
        if "dest" not in self:
            self[dest] = []
        self[dest].append(txt)

    def more(self, dest, nr=10):
        if "dest" in dir(self):
            res = self[dest][-nr:]
            del self[dest][-nr:]    
            return res
        else:
            return []

class Bot(Worker):

    """ Bot baseclass. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._sock = sys.stdout
        self.cache = OutputCache()
        self.cfg = Cfg(_default="")
        self.channels = []
        self.prompt = kwargs.get("prompt", True)
        self.verbose = kwargs.get("verbose", True)
   
    def announce(self, txt):
        """ announce to channel. """
        for channel in self.channels:
            self.say(channel, txt)

    def cmd(self, txt, options="", origin=""):
        """ run command from text. """
        self.start()
        e = Event()
        e.orig = repr(self)
        e.txt = txt
        e.options = options
        e.origin = origin or "root@shell"
        self.dispatch(e)
        e.wait()
        self.ready()
        return e

    def raw(self, txt):
        """ raw output on the socket. """
        if not self.verbose:
            return
        self._sock.write(str(txt))
        self._sock.write("\n")
        self._sock.flush()

    def say(self, channel, txt):
        """ convert channel output to txt. """
        self.cache.add(channel, txt)
        self.raw(txt)

    def start(self):
        fleet = Fleet()
        fleet.add(self)
        super().start()

    def wait_for_queue(self):
        """ wait for the queue to be empty. """
        while 1:
            time.sleep(0.01)
            s = self.queue.qsize()
            if s == 0:
                break

class Shell(Bot):

    """ Provides interface to a command line interface. """    

    def dispatch(self, event):
        """ display a prompt after dispatch. """
        super().dispatch(event)
        self.show_prompt()

    def get_event(self):
        e = Event()
        e.orig = repr(self)
        e.origin = "root@shell"
        e.txt = input()
        return e

    def show_prompt(self):
        """ show prompt. """
        if self.prompt:
            self._sock.write("> ")
            self._sock.flush()
