# OBJ - Framework to progam bots
#
# obj/handler.py
#
# Copyright 2019 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 01-01-2019 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

""" queued event handler. """

__version__ = 1

## imports

from obj.base import Dotted, Register, logdir
from obj.event import Event
from obj.fleet import Fleet
from obj.loader import Loader
from obj.utils import get_exception, get_run, launch, name

import _thread
import logging
import pkgutil
import queue
import socket
import threading
import types

## exceptions

class EMOD(Exception):

    """ module cannot be found. """

    pass

class ENOTIMPLEMENTED(Exception):

    """ base method is not inherited. """

    pass

class Handler(Loader):

    """ read event from blocking function and dispatch to overloaded function. """

    def __init__(self, *args, **kwargs):
        super().__init__()
        self._ready = threading.Event()
        self._stopped = False
        self._threaded = False
        self.cc = "!"
        self.queue = queue.Queue()

    def dispatch(self, event):
        """ parse the event and call corresponding command. """
        raise ENOTIMPLEMENTED

    def get_cmd(self, event):
        """ return commands. """
        cmd = event.get_cmd()
        return self.get_handler(cmd)

    def get_event(self):
        """ return event from queue. """
        return self.queue.get()

    def loop(self):
        """ the event loop. """
        while not self._stopped:
            event = self.get_event()
            if not event:
                break
            launch(self.dispatch, event)
        self.ready()

    def put(self, event):
        """ pass event over to queue. """
        self.queue.put_nowait(event)

    def ready(self):
        """ flag handler as ready. """
        self._ready.set()

    def register(self, cmd, callback):
        """ register a callback. """
        self._handlers[cmd] = callback

    def say(self, channel, txt):
        """ dummy print. """
        print(txt)

    def start(self):
        """ start the event loop and register this handler with the bus. """
        launch(self.loop)
         
    def stop(self): 
        """ stop this handler. """
        self._stopped = True
        self.queue.put(None)
        
    def wait(self):
        """ wait for handler to finish. """
        return self._ready.wait()

class EventHandler(Handler):

    """ provide dispatch to handle events. """

    def dispatch(self, event):
        """ dispatch event to exec. """
        event.parse()
        event._func = self.get_cmd(event)
        if event._func:
            logging.info("dispatch %s %s" % (event.origin, name(event._func)))
            try:
                event._func(event)
            except Exception as ex:
                logging.error(get_exception())
        event.show(self)
        event.ready()

class Worker(EventHandler):

    """ run from queue event handler. """

    def work(self):
        """ the event loop. """
        while not self._stopped:
            event = self.queue.get()
            if not event:
                break
            thr = launch(self.dispatch, event)
            thr.join()
        self.ready()
