# OBJ - Framework to program bots
#
# obj/loader.py
#
# Copyright 2019 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 01-01-2019 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

""" load modules into a table and scan for comands. """

__version__ = 1

## imports

from obj.base import Dotted, Register
from obj.fleet import Fleet
from obj.utils import get_exception, name, set_completer

import logging
import pkgutil
import importlib
import sys
import types

## classes

class Loader(Dotted):

    """ a loader can be used to load modules into a table (key, value pairs). """

    def __init__(self):
        super().__init__()
        self._table = Dotted()
        self._handlers = {}
        self._names = {}
    
    def direct(self, name, package=None):
        """ direct import of a module. """
        return importlib.import_module(name, package)

    def load_mod(self, modname, mod=None):
        """ register a module. """
        mod = mod or self.direct(modname)
        self._table[modname] = mod
        for key, func, modname in self.scan(mod):
            self._handlers[key] = func
            self._names[key] = modname
        return mod

    def get_handler(self, cmd):
        """ return handler registered. """
        val = self._handlers.get(cmd, None)
        if not val:
            modname = self._names.get(cmd, None)
            if modname:
                self.load_mod(modname)
            val = self._handlers.get(cmd, None)
        return val

    def register(self, cmd, handler):
        """ register a callback. """
        self._handlers[cmd] = handler

    def scan(self, mod):
        """ scan a module for commands. """
        for key in dir(mod):
            if key.startswith("_"):
                continue
            func = getattr(mod, key, None)
            if func and type(func) == types.FunctionType:
                if "event" in func.__code__.co_varnames:
                    yield (key, func, mod.__name__)

    def unload(self, modname):
        """ unload a module. """
        mod = self._table[modname]
        for key, func, name in self.scan(mod):
            if name == modname:
                del self._handlers[key]
                del self._names[key]

    def walk(self, modname, bot=None):
        """ browse a packages for all modules in it. """
        if not bot:
           bot = self
        try:
            mod = bot.load_mod(modname)
        except ModuleNotFoundError:
            logging.error("EMOD %s" % modname)
            return
        except Exception as ex:
            logging.error(get_exception())
            return
        if "__path__" in dir(mod):
            for nn in pkgutil.walk_packages(mod.__path__, modname+'.'):
                try:
                    bot.load_mod(nn[1])
                except ModuleNotFoundError:
                    logging.error("module %s not found" % nn)
                except Exception as ex:
                    logging.error(get_exception())
        return mod
