# OBJ - Framework to progam bots
#
# obj/utils.py
#
# Copyright 2019 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 01-01-2019 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

"""
    utility module.
    
"""

__version__ = 1

## imports

import atexit
import _thread
import datetime
import importlib
import logging
import logging.handlers
import obj.base
import optparse
import os
import re
import readline
import termios
import threading
import rlcompleter
import stat
import sys
import time
import traceback

## defines

opts = [
    ('-d', '', 'string', '', 'workdir', 'set working directory.'),
    ('-l', '', 'string', 'error', 'level', 'loglevel.'),
    ('-m', '', 'string', '', 'modules', 'modules to load.'),
    ('-o', '', "string", '', 'options', 'parseable options to use.'),
    ('', '--owner', "string", "", "owner", "userhost of the bot's owner"),
    ('', '--name', 'string', '', 'name', 'name of the program'),
    ('', '--text', 'string', '', 'txt', "text to display."),
    ('', '--progversion', 'string', '', 'version', "version of the program"),
] 

histfile = os.path.join(os.path.expanduser("~"), ".logs", "history")

timestrings = [
               "%a, %d %b %Y %H:%M:%S %z",
               "%d %b %Y %H:%M:%S %z",
               "%d %b %Y %H:%M:%S",
               "%a, %d %b %Y %H:%M:%S",
               "%d %b %a %H:%M:%S %Y %Z",
               "%d %b %a %H:%M:%S %Y %z",
               "%a %d %b %H:%M:%S %Y %z",
               "%a %b %d %H:%M:%S %Y",
               "%d %b %Y %H:%M:%S",
               "%a %b %d %H:%M:%S %Y",
               "%Y-%m-%d %H:%M:%S",
               "%Y-%m-%dt%H:%M:%S+00:00",
               "%a, %d %b %Y %H:%M:%S +0000",
               "%d %b %Y %H:%M:%S +0000",
              ]

# 2012-10-23 00:00:00
year_formats = [
    "%b %H:%M",
    "%b %H:%M:%S",
    "%a %H:%M %Y",
    "%a %H:%M",
    "%a %H:%M:%S",
    "%Y-%m-%d",
    "%d-%m-%Y",
    "%d-%m",
    "%m-%d",
    "%Y-%m-%d %H:%M:%S",
    "%d-%m-%Y %H:%M:%S",
    "%d-%m %H:%M:%S",
    "%m-%d %H:%M:%S",
    "%Y-%m-%d %H:%M",
    "%d-%m-%Y %H:%M",
    "%d-%m %H:%M",
    "%m-%d %H:%M",
    "%H:%M:%S",
    "%H:%M"
]

## instances

exceptions = []
resume = {}

## exceptions

class ENODATE(Exception):

    """ a date could not be determined. """

    pass

## classes

class Completer(rlcompleter.Completer):

    """ enables tab completion for the shell. """

    def __init__(self, commands=[], arguments=[]):
        super().__init__()
        self.commands = commands
        self.arguments = arguments

    def complete(self, text, state):
        """ suggest tab completion. """
        if state == 0:
            if text:
                self.matches = sorted([s for s in self.commands if s and s.startswith(text)])
            else:
                self.matches = self.commands[:]
            try:
                return self.matches[state]
            except IndexError:
                return None

    def display(self, substitution, matches, depth):
        """ display options to complete. """
        if matches:
            print("\n".join(matches))
            readline.redisplay()

class DumpHandler(logging.StreamHandler):

    """ A logging handler that ignores the log messages. """

    def emit(self, record):
        """ neglect logging record emission. """
        pass

## functions

def check_permissions(p, dirmask=0o700, filemask=0o600) -> None:
    """ check if directory permissions are set to readonly and if not set them to readonly. """
    from obj.base import cdir
    uid = os.getuid()
    gid = os.getgid()
    try:
        stats = os.stat(p)
    except FileNotFoundError:
        return
    except OSError:
        d, fn = os.path.split(p)
        cdir(d)
        stats = os.stat(d)
    if stats.st_uid != uid:
        os.chown(p, uid, gid)
    if os.path.isfile(p):
        mask = filemask
    else:
        mask = dirmask
    m = oct(stat.S_IMODE(stats.st_mode))
    if m != oct(mask):
        os.chmod(p, mask)

def close_history() -> None:
    """ close the history file (path is in histfile variable). """
    readline.write_history_file(histfile)

def construct(type, data={}, origin="", default=None):
    """ construct a type and initialize it with data JSON dict. """
    if origin:
        data["origin"] = origin
    cls = get_cls(type)
    if cls:
        try:
            return cls(data)
        except TypeError:
            return cls()

def daemon() -> None:
    """ switch to background mode. """
    pid = os.fork()
    if pid != 0:
        os._exit(0)
    os.chdir("/")
    os.setsid()
    os.umask(0)
    pid = os.fork()
    if pid != 0:
        os._exit(0)
    sys.argv[0] = "obj"
    sys.stdout.flush()
    sys.stderr.flush()
    si = open("/dev/null", 'r')
    so = open("/dev/null", 'a+')
    se = open("/dev/null", 'a+')
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())

def day():
     """" return the current day. """
     return str(datetime.datetime.today()).split()[0]

def days(o):
    """ show days till creation. """
    return elapsed(time.time() - fn_time(o._container.get("path", 0)))

def elapsed(seconds, short=True):
    """ show elapsed time. """
    txt = ""
    sub = str(seconds).split(".")[-1]
    nsec = float(seconds)
    year = 365*24*60*60
    week = 7*24*60*60
    day = 24*60*60
    hour = 60*60
    minute = 60
    years = int(nsec/year)
    nsec -= years*year
    weeks = int(nsec/week)
    nsec -= weeks*week
    days = int(nsec/day)
    nsec -= days*day
    hours = int(nsec/hour)
    nsec -= hours*hour
    minutes = int(nsec/minute)
    sec = nsec - minutes*minute
    if years:
        txt += "%sy" % years
    if weeks:
        days += weeks * 7
    if days:
        txt += "%sd" % days
    if years and short and txt:
        return txt
    if hours:
        txt += "%sh" % hours
    if days and short and txt:
        return txt
    if minutes:
        txt += "%sm" % minutes
    if hours and short and txt:
        return txt
    if sec == 0:
        txt += "0s"
    elif sec < 1 or not short:
        txt += "%.3fs" % sec
    else:
        txt += "%ss" % int(sec)
    txt = txt.strip()
    return txt

def enable_history() -> None:
    """" enable logging of history to the history file (path is in histfile variable). """
    from obj.base import cdir
    hf = histfile
    if not os.path.exists(hf):
        d, f = os.path.split(hf)
        cdir(d)
        touch(hf)
    readline.read_history_file(hf)
    atexit.register(close_history)

def fn_time(daystr: str) -> None:
    """ return the time determined from the filename of an object. """
    daystr = daystr.replace("_", ":")
    datestr = " ".join(daystr.split(os.sep)[-2:])
    datestr = datestr.split(".")[0]
    try:
        t = time.mktime(time.strptime(datestr, "%Y-%m-%d %H:%M:%S"))
    except ValueError:
        t = 0
    return t

def get_cls(name):
    try:
        modname, clsname = name.rsplit(".", 1)
        mod = importlib.import_module(modname)
        return getattr(mod, clsname)
    except:
        pass

def get_completer() -> Completer:
    """ return a completer. """
    return readline.get_completer()

def get_exception(txt="") -> str:
    """ return exception from stack. """
    exctype, excvalue, tb = sys.exc_info()
    trace = traceback.extract_tb(tb)
    result = ""
    for i in trace:
        fname = i[0]
        linenr = i[1]
        func = i[2]
        plugfile = fname[:-3].split(os.sep)
        mod = []
        for i in plugfile[::-1]:
            mod.append(i)
            if i == "obj":
                break
        ownname = '.'.join(mod[::-1])
        result += "%s:%s %s | " % (ownname, linenr, func)
    del trace
    res = "%s%s: %s %s" % (result, exctype, excvalue, str(txt))
    if res not in exceptions:
        exceptions.append(res)
    return res

def get_hour(daystr):
    """ get the hour from the string provided. """
    try:
        hmsre = re.search('(\d+):(\d+):(\d+)', str(daystr))
        hours = 60 * 60 * (int(hmsre.group(1)))
        hoursmin = hours  + int(hmsre.group(2)) * 60
        hms = hoursmin + int(hmsre.group(3))
    except AttributeError:
        pass
    except ValueError:
        pass
    try:
        hmre = re.search('(\d+):(\d+)', str(daystr))
        hours = 60 * 60 * (int(hmre.group(1)))
        hms = hours + int(hmre.group(2)) * 60
    except AttributeError:
        return 0
    except ValueError:
        return 0
    return hms

def list_mods(self, name):
    """ return list of modules in the obj package. """
    res = []
    loader = Loader()
    try:
        mod = loader.direct(name)
    except:
        logging.error(get_exception())
        return res
    if mod:
        if "__file__" in dir(mod):
            res.append(n)
        elif "__path__" in dir(mod):
            for nn in pkgutil.walk_packages(mod.__path__, name+'.'):
                res.append(nn[1])
    return res

def get_run():
    """ return runtime object. """
    from obj.base import Dotted, Store
    from obj.fleet import Fleet
    from obj.loader import Loader
    from obj.users import Users
    run = Dotted()
    run.aliases = Dotted(_default="")
    run.defaults = Dotted(_default="")
    run.fleet = Fleet()
    run.loader = Loader()
    run.store = Store()
    run.users = Users()
    return run

def get_time(daystr):
    """ return time with time.strptime. """
    for f in year_formats:
        try:
            t = time.mktime(time.strptime(daystr, f))
            return t
        except:
            pass

def level(level, logdir, nostream=False):
    """ set loglevel. """
    from obj.base import cdir
    if not os.path.exists(logdir):
        cdir(logdir)
    logfile = os.path.join(logdir, "rssbot.log")
    datefmt = '%H:%M:%S'
    format_time = "%(asctime)-8s %(message)-70s"
    format_plain = "%(message)-0s"
    level = level.upper()
    logger = logging.getLogger("")
    if logger.handlers:
        for handler in logger.handlers:
            logger.removeHandler(handler)
    if logger.handlers:
        for handler in logger.handlers:
            logger.removeHandler(handler)
    try:
        logger.setLevel(level)
    except ValueError:
        return
    formatter = logging.Formatter(format_plain, datefmt)
    if nostream:
        dhandler = DumpHandler()
        dhandler.propagate = False
        dhandler.setLevel(level)
        logger.addHandler(dhandler)
    else:
        ch = logging.StreamHandler()
        ch.propagate = False
        ch.setFormatter(formatter)
        ch.setLevel(level)
        logger.addHandler(ch)
    formatter2 = logging.Formatter(format_time, datefmt)
    filehandler = logging.handlers.TimedRotatingFileHandler(logfile, 'midnight')
    filehandler.propagate = False
    filehandler.setFormatter(formatter2)
    filehandler.setLevel(level)
    logger.addHandler(filehandler)
    return logger

def locked(func, *args, **kwargs):

    """ lock a method or function. """

    lock = _thread.allocate_lock()

    def lockedfunc(*args, **kwargs):
        """ encapsulate the function into a blocking one. """
        lock.acquire()
        res = None
        try:
            res = func(*args, **kwargs)
        finally:
            try:
                lock.release()
            except:
                pass
        return res
    return lockedfunc

def make_opts(options, usage, version) -> list:
    """ create commandline options. """
    parser = optparse.OptionParser(usage=usage, version=version)
    for option in options:
        optiontype, default, dest, helptype = option[2:]
        if "store" in optiontype:
            try:
                parser.add_option(option[0], option[1], action=optiontype, default=default, dest=dest, help=helptype)
            except Exception as ex:
                logging.error("%s option %s" % (str(ex), option))
                continue
        else:
            try:
                parser.add_option(option[0], option[1], type=optiontype, default=default, dest=dest, help=helptype)
            except Exception as ex:
                logging.error("^%s option %s" % (str(ex), option))
                continue
    args = parser.parse_args()
    return args

def name(obj):
    """ return a classname.methodname of a function. """
    try:
        n = "%s.%s" % (obj.__self__.__class__.__name__, obj.__name__)
    except:
        try:
            n = "%s.%s" % (obj.__class__.__name__, obj.__name__)
        except:
            try:
                n = obj.__class__.__name__
            except:
                n = obj.__name__
    return n

def now():
     """" return the current time. """
     return str(datetime.datetime.now()).split()[0]

def parse_cli(name, options=[], version=__version__, workdir=""):
    """ parse command line options. """
    from obj.base import cdir
    from obj.utils import opts
    usage = "%s [options]" % name
    ver = "%s #%s" % (name.upper(), version)
    opt, arguments = make_opts(options, usage, ver)
    from obj.base import cfg
    cfg.update(vars(opt))
    cfg.name = name.upper()
    cfg.version = version
    cfg.args = arguments
    if workdir:
        cfg.workdir = cfg.workdir or workdir
    if not cfg.workdir:
        cfg.workdir = os.path.abspath(os.path.join(os.path.expanduser("~"), ".obj"))
    if not os.path.exists(cfg.workdir):
        cdir(cfg.workdir)
    if cfg.args:
        cfg.shell = False
    else:
        cfg.shell = True
    if cfg.options:
        cfg.options = cfg.options.split(",")
    return cfg

def reset() -> None:
    """ reset terminal to previous state (prevents garblement on console). """
    close_history()
    if "old" in resume:
        try:
            termreset(resume["fd"], resume["old"])
        except:
            pass
            
def rtime():
    res =  str(datetime.datetime.now()).replace(" ", os.sep)
    return res

def scan_time(daystring):
    d = time.localtime()
    tstr = ""
    dstr = ""
    ystr = ""
    wday = ""
    mstr = ""
    for word in daystring.split():
        if word in weekdays:
            wday = word
        elif word in months:
            month = word
        elif "-" in word:
            dstr = word
        elif ":" in word:
            tstr = word
    if not tstr:
        tstr = str(d.tm_hour) + ":" + str(d.tm_min)
    if wday:
        difftonow = weekdays.index(wday) - d.tm_wday
        if difftonow < 0:
            difftonow = 7 + difftonow
        d = time.localtime(time.mktime(time.localtime()) + difftonow * 24* 60 * 60)
    if not dstr:
        dstr = str(d.tm_mday) + "-" + str(d.tm_mon) + "-" + str(d.tm_year)
    line = dstr + " " + tstr
    try:
        return get_time(line.strip())
    except ValueError:
        pass

def set_completer(commands, arguments=[]) -> None:
    """ set a list of options in the completer. """
    completer = Completer(commands, arguments)
    readline.set_completer_delims(' ')
    readline.parse_and_bind("tab: complete")
    readline.set_completion_display_matches_hook(completer.display)
    readline.set_completer(completer.complete)
    atexit.register(lambda: readline.set_completer(None))

def signal_term(sig, frame) -> None:
    """ handle the sighup signal, calls reload on all bus listeners. """
    for listener in bus.listeners:
         listener.reload()

def startup() -> None:
    """ run this function on startup, save the console state. """
    global resume
    enable_history()
    resume["fd"] = sys.stdin.fileno()
    resume["old"] = termsetup(sys.stdin.fileno())
    atexit.register(reset)

def stripped(jid):
    """ remove everything after a '/' """
    try:
        return str(jid).split("/")[0]
    except:
        return str(jid)

def termreset(fd, old) -> None:
    """ do an actual terminal reset, restores old state. """ 
    termios.tcsetattr(fd, termios.TCSADRAIN, old)

def termsetup(fd) -> list:
    """ return terminal setup. """
    old = termios.tcgetattr(fd)
    return old

def today():
    """ today at 0:00. """
    t = rtime().split(".")[0]
    ttime = time.strptime(t, "%Y-%m-%d/%H:%M:%S")
    result = time.mktime(ttime)
    return result

def to_day(daystring):
    previous = ""
    line = ""
    daystr = str(daystring)
    for word in daystr.split():
        if "-" in word:
            line += word + " "
        elif ":" in word:
            line += word
    if "-" not in line:
        line = day() + " " + line
    try:
        return get_time(line.strip())
    except ValueError:
        pass

def to_time(daystr):
    """ convert a string to a unix timestamp. """
    daystr = daystr.strip()
    if "," in daystr:
        daystr = " ".join(daystr.split(None)[1:7])
    elif "(" in daystr:
        daystr = " ".join(daystr.split(None)[:-1])
    else:
        try:
            d, h = daystr.split("T")
            h = h[:7]
            daystr = " ".join([d, h])
        except:
            pass
    res = 0
    for tstring in timestrings:
        try:
            res = time.mktime(time.strptime(daystr, tstring))
        except:
            try:
                res = time.mktime(time.strptime(" ".join(daystr.split()[:-1]), tstring))
            except:
                pass
        if res:
           break
    if not res:
        raise ENODATE(daystr)
    return res

def touch(fname) -> None:
    """ touch a file, creates a file without any content. """
    try:
        fd = os.open(fname, os.O_RDONLY | os.O_CREAT)
        os.close(fd)
    except TypeError:
        pass

def launch(func, *args, **kwargs):
    """ launch tasks. """
    from obj.task import Task
    t = Task()
    t.put(func, *args, **kwargs)
    t.start()
    return t

def kill(thrname=""):
    """ kill tasks. """
    for task in running(thrname):
        if "cancel" in dir(task):
            task.cancel()
        if "exit" in dir(task):
            task.exit()
        if "stop" in dir(task):
            task.stop()

def running(tname=""):
    """ show running tasks. """
    for task in threading.enumerate():
        name = str(task)
        if name.startswith("<_"):
            continue
        if tname and tname.upper() not in name.upper():
            continue
        yield task

